#grid {
    margin-top: 4;
    layout: grid;
    grid-size: 2 3;
}

Grid > Container {
    padding: 2 4;
    height: 1fr;
}

.dropdown {
    max-width: 36;
    max-height: 4;
}

#dropdown-2 {
    background: $accent;
    scrollbar-color: $secondary;
}

#dropdown-2 .autocomplete--highlight-match {
    color: $text;
    background: $secondary;
}
#dropdown-2 .autocomplete--selection-cursor {
    background: $accent-lighten-1;
}

#dropdown-2 .autocomplete--left-column {
    background: $accent-darken-1;
}

#dropdown-2 .autocomplete--right-column {
    background: $accent-darken-2;
}

#dropdown-3 {
    border-bottom: wide $success;
    scrollbar-color: $success;
}

#dropdown-3 .autocomplete--highlight-match {
    color: $text;
    background: $success;
}

#dropdown-3 .autocomplete--selection-cursor {
    background: $success 20%;
}

#dropdown-4 {
    scrollbar-color: deeppink;
}

#dropdown-4 .autocomplete--highlight-match {
    color: $text;
    background: deeppink 60%;
    text-style: not bold italic underline;
}

#dropdown-4 .autocomplete--selection-cursor {
    background: deeppink 20%;
}

#input-4  {
    border: tall deeppink;
}

#input-4>.input--cursor {
    color: deeppink;
}