Screen {
    /*layers: textual-autocomplete;*/
}

#search-box {
    width: 75%;
}

#lead-text {
    width: 100%;
    content-align: center middle;
    color: $secondary;
    text-style: bold;
    padding: 1 1 1 0;
}

#my-dropdown {
    max-width: 40;
    max-height: 6;
    /*border-left: wide magenta;*/
    /*scrollbar-color: greenyellow;*/
    /*background: darkgreen;*/
}

/*#my-dropdown .autocomplete--highlight-match {*/
/*    color: yellow;*/
/*    background: magenta;*/
/*    text-style: bold;*/
/*}*/

/*#my-dropdown .autocomplete--cursor {*/
/*    background: $boost;*/
/*}*/

/*#my-dropdown .autocomplete--left-column {*/
/*    background: mediumpurple;*/
/*}*/


/*#my-dropdown .autocomplete--main-column {*/
/*    background: deeppink;*/
/*}*/

/*#my-dropdown .autocomplete--right-column {*/
/*    background: lightseagreen;*/
/*}*/

#info-text {
    padding: 1 8;
    color: $text;
    height: auto;
    width: 100%;
    text-align: center;
}

